/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.net;

import java.io.IOException;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.InetSocketAddress;
import java.net.MulticastSocket;
import java.net.NetworkInterface;
import java.net.SocketAddress;

import casa.util.Logger;
import casa.dodwan.util.Bufferizer;
import casa.util.DataBufferizer;
import casa.dodwan.config.DodwanEnvironment;

public class MulticastInetSocket<DOCTYPE> extends InetSocket<DOCTYPE> {
    
    private int IPTOS_RELIABILITY = 0x4;
    
    private int IPTOS_THROUGHPUT = 0x8;
    
    private int IPTOS_LOWDELAY = 0x10;
    
    InetAddress mcastAddress_ = null;
    InetAddress ifAddress_ = null;
    
    private long lastJoin_ = 0;
    private long joinThreshold_ = 60;
    private long echoThreshold_ = 90;

    // ------------------------------------------------------------
    public MulticastInetSocket(InetSocketAddress localSocketAddress,
			       InetAddress mcastAddress, 
			       InetAddress ifAddress,
			       Bufferizer<DOCTYPE> bufferizer)
	throws Exception {
	
	//  	    verbosity.enable();
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket(" + localSocketAddress
			       + ", " + mcastAddress
			       + ", " + ifAddress
			       + ", " + bufferizer + ")");
	
	long now = System.currentTimeMillis();
	lastEcho = now;
	lastJoin_ = now;
	
	init(localSocketAddress, mcastAddress, ifAddress, bufferizer);
    }
    
    // ------------------------------------------------------------
    public InetAddress getInterface() throws Exception {
	
	return ((MulticastSocket) socket_).getInterface();
	
    }
    
    // ------------------------------------------------------------
    public NetworkInterface getNetworkInterface() throws Exception {
	
	return ((MulticastSocket) socket_).getNetworkInterface();
	
    }
    
    // ------------------------------------------------------------
    public void init(InetSocketAddress localSocketAddress,
		     InetAddress mcastAddress, 
		     InetAddress ifAddress,
		     Bufferizer<DOCTYPE> bufferizer) throws Exception {
	
	bufferizer_ = bufferizer;
	localSocketAddress_ = localSocketAddress;
	mcastAddress_ = mcastAddress;
	ifAddress_ = ifAddress;

	if (ifAddress != null) {
	    NetworkInterface nif = NetworkInterface.getByInetAddress(ifAddress);
	    if (nif == null) {
		Logger.log("socket", "warning: interface " + ifAddress + " does not exist");
	    }
	}

	MulticastSocket msocket = new MulticastSocket(localSocketAddress_);
	socket_ = msocket;
	
	// FG: not sure this is effective...
	socket_.setReceiveBufferSize(5242880);

	// msocket.setLoopbackMode(true);

	Logger.log("socket", "local=" + localSocketAddress_);
	Logger.log("socket", "mcast=" + mcastAddress_);
	if (ifAddress != null)
	    Logger.log("socket", "iface=" + ifAddress_);
	Logger.log("socket", "cookie=" + cookie_
		   + ",rcvBufferSize=" + socket_.getReceiveBufferSize());
	Logger.log("socket", "loopbackMode=" + msocket.getLoopbackMode()
		   + ",echoThreshold=" + echoThreshold_ + " s"
		   + ",joinThreshold=" + joinThreshold_ + " s");

	// Join muticast group for the first time
	if (ifAddress == null) {
	    // System.out.println("Joining group " + mcastAddress_
	    // 		       + " on default interface");
	    joinGroup(mcastAddress_);
	}
	else {
	    // System.out.println("Joining group " + mcastAddress_
	    // 		       + " on interface "
	    // 		       + ifAddress_);
	    joinGroup(mcastAddress_,
		      ifAddress_);
	}
    }
    
    // ------------------------------------------------------------
    protected void sendBytes(InetAddress addr, int port,
			     byte[] buffer, int offset, int length)
	throws IOException {
	
	if (echoAddress != null) 
	    checkLastEcho();
	
	super.sendBytes(addr, port, buffer, offset, length);
    }
    
    // ------------------------------------------------------------
    private void checkLastEcho() {
	
	long now = System.currentTimeMillis();
	
	// If lastEcho is recent enough, there is nothing to do...
	if (now - lastEcho < (1000 * echoThreshold_))
	    return;
	
	// If a multicast subscription has been attempted recently,
	// then there is no need to attempt another one...
	if (now - lastJoin_ < (1000 * joinThreshold_))
	    return;
	
	// OK, then... Let's try to join the multicast group again!
	Logger.log("socket", "Last echo was " 
		   + ((now - lastEcho) / 1000) 
		   + " seconds ago...");
	Logger.log("socket", "Attempting to join multicast group again");
	try {
	    joinAgain();
	}
	catch (Exception e) {
	    System.out.println("mcast_socket: " + e);
	}
    }
    
    
    // ------------------------------------------------------------
    public void joinGroup(InetAddress mcastAddress) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket.joinGroup(" + mcastAddress
			       + ")");
	mcastAddress_ = mcastAddress;
	ifAddress_ = null;
	
	Logger.log("socket", "join " + mcastAddress);
	lastJoin_ = System.currentTimeMillis();
	
	((MulticastSocket) socket_).joinGroup(mcastAddress);
	
	//try {
	//	((MulticastSocket) socket_).setTrafficClass(IPTOS_RELIABILITY);
	//}
	//catch (Exception e) {
	//	e.printStackTrace();
	//}
    }
    
    // ------------------------------------------------------------
    public void joinGroup(InetAddress mcastAddress, 
			  InetAddress ifAddress)
	throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket.joinGroup(" + mcastAddress
			       + ", " + ifAddress + ")");
	
	mcastAddress_ = mcastAddress;
	ifAddress_ = ifAddress;
	
	Logger.log("socket", "join " + mcastAddress + " on interface " + ifAddress);
	
	NetworkInterface nif = NetworkInterface.getByInetAddress(ifAddress);
	if (nif == null) {
	    Logger.log("socket", "interface " + ifAddress + " does not exist");
	    return;
	}
	lastJoin_ = System.currentTimeMillis();
	
	((MulticastSocket) socket_).joinGroup(new InetSocketAddress(mcastAddress, 0), nif);

    }
    
    // ------------------------------------------------------------
    public void joinAgain() throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket.joinAgain()");
	
	if (ifAddress_ == null)
	    joinGroup(mcastAddress_);
	else 
	    joinGroup(mcastAddress_, ifAddress_);
    }
    
    // ------------------------------------------------------------
    public void leaveGroup(InetAddress mcastAddress) throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket.leaveGroup(" + mcastAddress
			       + ")");
	
	((MulticastSocket) socket_).leaveGroup(mcastAddress);
    }
    
    // ------------------------------------------------------------
    public void leaveGroup(SocketAddress mcastAddress, NetworkInterface netIf)
	throws Exception {
	
	if (verbosity.isEnabled())
	    System.out.println("MulticastInetSocket.leaveGroup(" + mcastAddress
			       + ", " + netIf + ")");
	
	((MulticastSocket) socket_).leaveGroup(mcastAddress, netIf);
    }
    
    // ------------------------------------------------------------
    public void setNetworkInterface(NetworkInterface iface) throws Exception {
 
      	// YM : Sans doute pas utilisable par DoDWAN en l'état :
    	//  DoDWAN ne prevoit pas de désigner une interface autrement que par une adresse
    	// (parametre interface_address)
    	// 
    	((MulticastSocket) socket_).setNetworkInterface(iface);
    }


    // ------------------------------------------------------------
    public void setInterface(InetAddress inf) throws Exception {

    	// YM memorise la nouvelle interface (pour un futur join)
    	ifAddress_ = inf;

    	Logger.log("socket", "set interface " + inf);

    	((MulticastSocket) socket_).setInterface(inf);
    }

    // ------------------------------------------------------------
    public void setTTL(int ttl) {
	try {
	    ((MulticastSocket) socket_).setTimeToLive(ttl);
	}
	catch (IOException e) {
	    e.printStackTrace();
	}
    }
    
    public int getTTL() {
	int ttl = 0;
	try {
	    ttl = ((MulticastSocket) socket_).getTimeToLive();
	}
	catch (IOException e) {
	    e.printStackTrace();
	}
	return ttl;
    }
    
    // ------------------------------------------------------------
    @Override
	public String toString() {
	String result = "MulticastInetSocket("
	    + localSocketAddress_ + ")";
	return result;
    }
}
