/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/


package casa.dodwan.net;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.util.Set;
import java.util.HashSet;
import java.net.SocketAddress;
import java.net.InetSocketAddress;
import java.net.Socket;
import java.net.SocketTimeoutException;

import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.util.Logger;
import casa.util.DataBufferizer;
import casa.dodwan.util.SystemEnvironment;
import casa.dodwan.util.DataFormatting;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageBufferizer;

public class TCP_Channel 
    extends Thread 
    implements Channel {

    public static short COOKIE = 13579;

    private Socket socket_ = null;

    private InetSocketAddress localSocketAddr_;
    private InetSocketAddress remoteSocketAddr_;
    private int timeout_;
    private MessageBufferizer out_bufferizer_ = 
	new MessageBufferizer();
    private MessageBufferizer in_bufferizer_ = 
	new MessageBufferizer();

    private OutputStream out_ = null;
    private InputStream  in_  = null;

    private long period_ = 0;
    private BeaconSender sender_ = null;
    private ChannelManager manager_;
    private boolean isOpen_ = false;

    private String peerId_ = null;
    private Set<String> peers_ = new HashSet<String>();

    private int bufSize_ = 0;
    
    private long lastRx_ = System.currentTimeMillis();
    
    // ============================================================
    public TCP_Channel(Socket socket,
		       ChannelManager manager) {

	setDaemon(true);
	
	localSocketAddr_  = (InetSocketAddress)socket.getLocalSocketAddress();
	remoteSocketAddr_ = (InetSocketAddress)socket.getRemoteSocketAddress();

	Logger.log("tcp_chan", "accept " + this);

	socket_ = socket;
	manager_ = manager;
	period_ = 15;
	
	sender_ = new BeaconSender(this, period_ * 1000);
	sender_.enable();
    }
    
    // ------------------------------------------------------------
    public static String getSyntax() {

	return "<-ra remoteAddr> <-rp remotePort> [-la localAddr] |-lp localPort] [-to timeout] [-bs bufSize]";
    }

    // ------------------------------------------------------------
    public TCP_Channel(String[] arguments,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;

	String localAddr  = null;
	String remoteAddr = null;
	int localPort  = 0;
    	int remotePort = 0;
	int timeout = 2000;
	int bufSize = 0;
	
	remoteAddr = ArgumentParsing.searchString("-ra", arguments);
	remotePort = ArgumentParsing.searchInt("-rp", arguments);
	try {
	    localAddr = ArgumentParsing.searchString("-la", arguments);
	}
	catch (ArgumentNotFoundException e) {}
	try {
	    localPort = ArgumentParsing.searchInt("-lp", arguments);
	}
	catch (ArgumentNotFoundException e) {}
	try {
	    timeout = ArgumentParsing.searchInt("-to", arguments);
	}
	catch (ArgumentNotFoundException e) {}
	try {
	    bufSize = ArgumentParsing.searchInt("-bs", arguments);
	}
	catch (ArgumentNotFoundException e) {}

	InetSocketAddress locAddr = null;
	if ((localAddr != null) || (localPort > 0))
	    locAddr = new InetSocketAddress(localAddr, localPort);
	    
	InetSocketAddress remAddr =
	    new InetSocketAddress(remoteAddr, remotePort);

	init(locAddr, remAddr, timeout, bufSize);
    }
    
    // ------------------------------------------------------------
    public TCP_Channel(String remoteAddr,
		       int remotePort,
		       int timeout,
		       int bufSize,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;

	InetSocketAddress remAddr =
	    new InetSocketAddress(remoteAddr, remotePort);

	init(null, remAddr, timeout, bufSize);
    }

    // ------------------------------------------------------------
    public TCP_Channel(InetSocketAddress remoteSocketAddr, 
		       int timeout,
		       int bufSize,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;

	init(null, remoteSocketAddr, timeout, bufSize);
    }

    // ------------------------------------------------------------
    public TCP_Channel(InetSocketAddress localSocketAddr,
		       InetSocketAddress remoteSocketAddr, 
		       int timeout,
		       int bufSize,
		       ChannelManager manager)
	throws Exception {

	manager_ = manager;

	init(localSocketAddr, remoteSocketAddr, timeout, bufSize);
    }

    // ------------------------------------------------------------
    private void init(InetSocketAddress localSocketAddr,
		      InetSocketAddress remoteSocketAddr, 
		      int timeout,
		      int bufSize)
	throws Exception {

	localSocketAddr_  = localSocketAddr;
	remoteSocketAddr_ = remoteSocketAddr;
	timeout_ = timeout;
	period_ = 15;
	bufSize_ = bufSize;
	
	Logger.log("tcp_chan", "connect " + this);

	openConnection();

	sender_ = new BeaconSender(this, period_ * 1000);
	sender_.enable();
    }

    // ------------------------------------------------------------
    public InetSocketAddress localAddress() {

	return (InetSocketAddress)socket_.getLocalSocketAddress();
    }

    // ------------------------------------------------------------
    public InetSocketAddress remoteAddress() {

	return (InetSocketAddress)socket_.getRemoteSocketAddress();
    }

    // ------------------------------------------------------------
    public Set<String> getPeers() {

	return peers_;
    }

    // ------------------------------------------------------------
    public void run() {

	if (socket_ == null) {
	    try {
		openConnection();
	    }
	    catch (Exception e) {
		return;
	    }
	    if (socket_ == null)
		return;
	}

	try {
	    out_ = socket_.getOutputStream();
	    in_  = socket_.getInputStream();
	    
	    handshake();
	    if (peerId_ == null) {
		close();
		return;
	    }

	    peers_.add(peerId_);
	    isOpen_ = true;

	    if (manager_ != null)
		manager_.peerDiscovered(peerId_, this);

	    // FG
	    // sender_.resume();

	    while (true) {
		byte[] buffer = Stream_IO.receiveBundle(in_);
		// Logger.log("tcp_chan", "< [" + buffer.length + "] via " + this);
		lastRx_ = System.currentTimeMillis();

		Message msg = in_bufferizer_.fromBuffer(buffer);

		if (manager_ != null)
		    manager_.process(msg);
	    }
	}
	catch(Exception e) {
	    if (isOpen_)
		Logger.log("tcp_chan", "rcv_failure on " 
			   + this);
	}

	// Logger.log("tcp_chan", "terminate " 
	// 	   + this);

	if (isOpen_)
	    close();
    }
    
    // ------------------------------------------------------------
    public void open() {
	
	Logger.log("tcp_chan", "open "
		   + this);
	
	start();
    }
    
    // ------------------------------------------------------------
    private void openConnection()
	throws Exception {

	try {
	    socket_ = new Socket();
	    if (bufSize_ > 0) {
		socket_.setSendBufferSize(bufSize_);
		socket_.setReceiveBufferSize(bufSize_);
	    }
	    if (localSocketAddr_ != null)
		socket_.bind(localSocketAddr_);
	    socket_.connect(remoteSocketAddr_, timeout_);
	    InetSocketAddress localAddr =
		(InetSocketAddress)socket_.getLocalSocketAddress();
	    localSocketAddr_ = new InetSocketAddress(localAddr.getAddress(),
						     localAddr.getPort());
	    
	    // socket_.setSoTimeout(timeout_);
	    // Logger.log("tcp_chan", "open " + this);
	}
	catch (Exception e) {
	    Logger.log("tcp_chan", "open failed ("
		       + e + ")");
	    throw e;
	}
    }

    // ------------------------------------------------------------
    private void handshake() 
	throws Exception {

	// Let us first send a 2-byte cookie to the remote peer
	byte[] buffer = new byte[2];
	DataBufferizer.writeShort(COOKIE, buffer, 0);
	Stream_IO.sendBytes(out_, buffer);
	
	// Logger.log("tcp_chan", "> cookie=" 
	// 	   + COOKIE);
	
	// Let us now send our id to the remote peer
	String id = SystemEnvironment.host;
	buffer = new byte[id.length() + 2];
	DataBufferizer.writeString(id, buffer, 0);
	Stream_IO.sendBundle(out_, buffer);
	
	// Logger.log("tcp_chan", "> id="
	// 	   + id);
	
	// Let us now receive a 2-byte cookie from the remote peer
	buffer = Stream_IO.receiveBytes(in_, 2);
	short cookie = DataBufferizer.readShort(buffer, 0);
	boolean cookie_ok = (cookie == COOKIE);

	// Logger.log("tcp_chan", "< cookie="
	// 	   + cookie
	// 	   + " ("
	// 	   + (cookie_ok?"ok":"invalid")
	// 	   + ")");

	if (! cookie_ok) {
	    Logger.log("tcp_chan", "warning: handshake failed");
	    return;
	}

	// Let us now receive the remote peer's id
	buffer = Stream_IO.receiveBundle(in_);
	lastRx_ = System.currentTimeMillis();
	peerId_ = DataBufferizer.readString(buffer, 0);
	Logger.log("tcp_chan", "handshake ok (peer=" + peerId_ + ")");
    }

    // ------------------------------------------------------------
    public void close() {

	if (! isOpen_)
	    return;

	Logger.log("tcp_chan", "close " + this);
	isOpen_ = false;
	
	if (sender_ != null)
	    sender_.suspend();
	
	try {
	    if (! socket_.isClosed()) {
		socket_.close();
	    }
	}
	catch (Exception e) {}

	if (manager_ != null) {
	    manager_.peerLost(peerId_);
	    manager_.channelClosed(this);
	}
    }

    // ------------------------------------------------------------
    public boolean isOpen() {
	return isOpen_;
    }

    // ------------------------------------------------------------
    public void write(Message msg) {

	if (! isOpen_ ) {
	    Logger.log("tcp_chan", "warning: channel closed");
	    return;
	}

	try {
	    synchronized(this) {

		byte[] buffer = out_bufferizer_.toBuffer(msg);

		// Logger.log("tcp_chan", "> [" + buffer.length + "] rcv="
		//  	   + peerId_ + " via " + this);

		StreamSender sender = new StreamSender(out_, buffer);
		sender.start();
	    }
	}
	catch (Exception e) {
	    Logger.log("tcp_chan", "warning: tx failed ("
		       + this + ")");
	}
    }

    // ------------------------------------------------------------
    public String toString() {

	String buf = null;
	try {
	    buf = ",buf=" + socket_.getSendBufferSize()
		+ "/" + socket_.getReceiveBufferSize();
	}
	catch (Exception e) {}
	
	return "TCP_Channel(local="
	    + DataFormatting.toString(localSocketAddr_)
	    + ",remote="
	    + DataFormatting.toString(remoteSocketAddr_)
	    + ",timeout=" + timeout_
	    + (buf!=null?buf:"")
	    + ")";
    }

    // ------------------------------------------------------------
    public long lastRx() {
	return lastRx_;
    }
}
