/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.pubsub;

import java.io.File;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Set;
import java.net.InetSocketAddress;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageReceiver;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.Command;
import casa.dodwan.util.Console;
import casa.dodwan.util.DataFormatting;
import casa.util.Processor;


// import masc.metier.DocumentForwarder;

public class PubSubServiceConsole extends Console {
    
    
    public PubSubService pubSubService_;
    
    // ------------------------------------------------------------
    public PubSubServiceConsole(PubSubService pubSubService) {
	
	pubSubService_ = pubSubService;
	prompt_ = "pubsub% ";
    }
    
    // ============================================================
    // MAIN COMMANDS
    // ============================================================
    
    // ------------------------------------------------------------
    public PubSubServiceConsole(String prompt, PubSubService pubSubService) {
	
	pubSubService_ = pubSubService;
	prompt_ = prompt;
    }
    

    // ------------------------------------------------------------
    // publish a message
    public void publish(String[] arguments, PrintStream out) {
	String descString = null;
	String msg = null;
	int size = -1;
	String fname = null;
	try {
	    descString = ArgumentParsing.searchString("-desc", arguments);
	    try {
		msg =  ArgumentParsing.searchString("-msg", arguments);
	    }
	    catch (ArgumentNotFoundException e) {}
	    try {
		fname = ArgumentParsing.searchString("-f", arguments);
	    }
	    catch (ArgumentNotFoundException e) {}
	    try {
		size = ArgumentParsing.searchInt("-s", arguments);
	    }
	    catch (ArgumentNotFoundException e) {}
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	    out.println("Error: syntax error");
	}
	
	if ((msg == null) && (fname == null) && (size == -1)) {
	    out.println("Missing argument -msg or -file or -size");
	    return;
	}
	
	Descriptor desc = Descriptor.fromAttributes(descString);
	String docid = desc.getDocumentId();
	if (docid == null)
	    desc.setUniqueDocumentId();

	if (msg != null) {
	    try {
		msg += "\n";
		pubSubService_.publish(desc, msg.getBytes());
		out.println("Done");
	    } catch (Exception e) {
		out.println(e);
		out.println("Error: failed to publich message");
	    } 
	    return;
	}

	if (size != -1) {
	    try {
		byte[] buffer = new byte[size];
		pubSubService_.publish(desc, buffer);
		out.println("Done");
	    } catch (Exception e) {
		out.println(e);
		out.println("Error: failed to publich message");
	    } 
	    return;
	}

	    
	if (fname != null) {
	    try {
		File file = new File(fname);
		if (! file.exists()) {
		    out.println("File does not exist");
		    return;
		}
		desc.setAttribute("file", file.getName());
		pubSubService_.publish(desc, file);
		out.println("Done");
	    } catch (Exception e) {
		out.println(e);
		out.println("Error: failed to publich message");
	    } 
	    return;
	}
    }
	


    // ------------------------------------------------------------
    // add a subscription
    public void add(String[] arguments, PrintStream out) {
	
	String key = null;
	InetSocketAddress addr = null;
	String path = null;
	String cmd = null;
	String descString = null;
	
	try {
	    key = ArgumentParsing.searchString("-k", arguments);
	    descString = ArgumentParsing.searchString("-desc", arguments);
	    
	    try {
		String address = ArgumentParsing.searchString("-a", arguments);
		addr = DataFormatting.toInetSocketAddress(address);
	    } catch (ArgumentNotFoundException e) {
	    }
	    
	    try {
		path = ArgumentParsing.searchString("-d", arguments);
	    } catch (ArgumentNotFoundException e) {
	    }
	    
	    try {
		cmd = ArgumentParsing.searchString("-e", arguments);
	    } catch (ArgumentNotFoundException e) {
	    }
	    
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	    out.println("Error: syntax error");
	}

	if (pubSubService_.getReceiver(key) != null) {
	    out.println("Error: a subscription with key " + key + " already exists");
	    return;
	}
	
	Descriptor pattern = null;
	if (descString.equals("ANY"))
	    pattern = new Descriptor();
	else
	    pattern = Descriptor.fromAttributes(descString);
	Processor<Descriptor> processor = null;

	if (path != null) {
	    File dir = new File(path);
	    if (! dir.isDirectory()) {
		out.println("Error: invalid directory " + path);
		return;
	    }
	    processor = new MessageSaver(dir, pubSubService_);
	} else if (cmd != null) {
	    File file = new File(cmd);
	    if (! file.canExecute()) {
		out.println("Error: invalid command " + cmd);
		return;
	    }
	    processor = new MessagePiper(cmd, pubSubService_);
	} else if (addr != null) {
	    processor = new MessageForwarder(addr, pubSubService_);
	}
	pubSubService_.addSubscription(key, pattern, processor);
	out.println("Done");
    }
    
    // ------------------------------------------------------------
    // save message to file
    public void save(String[] arguments, PrintStream out) {
	
	String key = null;
	String file = null;
	
	try {
	    key = ArgumentParsing.searchString("-k", arguments);
	    file = ArgumentParsing.searchString("-f", arguments);
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}

	try {
	    pubSubService_.getAsFile(key, new java.io.File(file));
	}
	catch (Exception e) {
	    out.println(e);
	}
    }

    // ------------------------------------------------------------
    public void disable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	pubSubService_.disable();
    }
    
    // ------------------------------------------------------------
    public void enable(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	pubSubService_.enable();
    }
    
    // ------------------------------------------------------------
    @Override
	public void help(String offset, PrintStream out) {
	out
	    .println(offset
		     + "[e]nable / [d]isable\n"
		     + offset
		     + "[su]spend / [res]ume\n"
		     + offset
		     + "[st]atus\n"
		     + offset
		     + "[v]erbosity <level>\n"
		     + offset
		     + "[a]dd [-k key]  [-desc desc_string] <-d path|-e cmd|-a addr:port>\n"
		     + offset
		     + "[p]ublish  <-desc desc_string> <-f file|-msg string|-s size>\n"
		     + offset + "[rem]ove [keys...]>\n" 
		     + offset + "[sa]ve [-k key] [-f file]>\n" 
		     + offset + "[l]ist\n" 
		     + offset + "[k]eys\n"
		     + offset + "[c]lear\n"
		     + offset + "[sh]ow <key>+");
    }
    
    // -------------------------------------------------------------
    public void processCommand(String[] arguments, BufferedReader reader,
			       PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("a")) {
	    add(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("c")) {
	    clear(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("p")) {
	    publish(arguments, out);
	    return;
	}

	if (arguments[0].startsWith("rem")) {
	    remove(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("d")) {
	    disable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("e")) {
	    enable(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("h")) {
	    out.println("Pub/Sub Service: Command Set\n");
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("k")) {
	    keys(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("v")) {
	    verbosity(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("res")) {
	    resume(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("sa")) {
	    save(arguments, out);
	    return;
	}

	if (arguments[0].startsWith("sh")) {
	    show(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("st")) {
	    status(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("su")) {
	    suspend(arguments, out);
	    return;
	}
	
	out.println("Unknown command");
	help("  ", out);
    }
    
    // ------------------------------------------------------------
    public void remove(String[] arguments, PrintStream out) {
	
	if (arguments.length == 1) {
	    help("  ", out);
	    return;
	}
	
	for (int i = 1; i < arguments.length; i++) 
	    pubSubService_.removeSubscription(arguments[i]);

	out.println("Done");
    }
    
    // ------------------------------------------------------------
    public void resume(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	pubSubService_.resume();
    }
    
    // ------------------------------------------------------------
    public void keys(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("\nSubscription keys:\n");
	Set keys = pubSubService_.getKeys();
	out.println(DataFormatting.toString(keys, "\n"));
    }
    
    // ------------------------------------------------------------
    public void clear(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("\nClearing subscriptions\n");
	pubSubService_.clear();
    }
    
    // ------------------------------------------------------------
    private void showReceiver(MessageReceiver recv, 
			      PrintStream out) {

	out.println("  Pattern:   " + recv.getPattern());
	out.println("  Processor: " + recv.getProcessor());
    }

    // ------------------------------------------------------------
    public void show(String[] arguments, PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	out.println("show subscription(s)");
	
	if (arguments.length == 1) 
	    for (String key: pubSubService_.getKeys()) {
		MessageReceiver recv = 
		    pubSubService_.getReceiver(key);
		if (recv != null)
		    showReceiver(recv, out);
	    }
	else
	    for (int i = 1; i < arguments.length; i++) {
		MessageReceiver recv = 
		    pubSubService_.getReceiver(arguments[i]);
		if (recv != null)
		    showReceiver(recv, out);
	    }
    }
    
    // ------------------------------------------------------------
    public void source(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	parseCommandFile(arguments[0], out);
    }
    
    // ============================================================
    // COMMAND PARSING
    // ============================================================
    
    // ------------------------------------------------------------
    public void status(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println(pubSubService_);
    }
    
    
    // ------------------------------------------------------------
    public void suspend(String[] arguments, PrintStream out) {
	
	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	pubSubService_.suspend();
    }
    
    // ------------------------------------------------------------
    public void verbosity(String[] arguments, PrintStream out) {
	
	if (arguments.length != 2) {
	    help("  ", out);
	    return;
	}
	
	int level = -1;
	try {
	    level = Integer.parseInt(arguments[1]);
	} catch (Exception e) {
	    out.println("Could not parse verbosity level");
	}
	
	if (level >= 0)
	    pubSubService_.setVerbosity(level);
    }
    
}
