/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.queue;

import java.util.List;
import java.util.LinkedList;
import java.io.File;
import java.io.FileReader;
import java.io.PrintStream;
import java.io.FileOutputStream;
import java.io.BufferedReader;

import casa.dodwan.docware.Descriptor;
import casa.dodwan.docware.DescriptorZipBufferizer;
import casa.dodwan.docware.Payload;
import casa.dodwan.docware.PayloadBufferizer;
import casa.dodwan.message.Message;
import casa.util.Processor;
import casa.dodwan.util.Repository;
import casa.dodwan.util.MemoryBasedRepository;
import casa.dodwan.util.FileBasedRepository;
import casa.dodwan.pubsub.PubSubService;

public class MessageQueue 
    implements Processor<Descriptor> {
    
    private Object lock_ = new Object();

    private File path_;
    
    // Persistent map containing all descriptors
    private Repository<String, Descriptor> descriptors_;
    
    // Persistent map containing all payloads
    private Repository<String, Payload> payloads_;
    
    private List<String> keys_ = 
	new LinkedList<String>();

    private PrintStream fout_ = null;

    private PubSubService pubSubService_;

    // ------------------------------------------------------------
    public MessageQueue(File path, PubSubService pubSubService) {

	path_ = path;
	pubSubService_ = pubSubService;

	if (path_ == null) {
	    descriptors_ = new MemoryBasedRepository<String, Descriptor>();
	    payloads_    = new MemoryBasedRepository<String, Payload>();
	}
	else {
	    try {
	    
		if (!path_.exists())
		    path_.mkdirs();
		
		descriptors_ = 
		    new FileBasedRepository<Descriptor>(new File(path_, "descriptors"), 
							new DescriptorZipBufferizer());
		payloads_ =
		    new FileBasedRepository<Payload>(new File(path_, "payloads"), 
						 new PayloadBufferizer());
	    
		// Prepare history file
		File file = new File(path_, "history");
		File bakFile = new File(file.getAbsoluteFile() + ".bak");
	    
//		if (! file.exists()) {
		    fout_ = new PrintStream(new FileOutputStream(file,true),true);
//		    return;
//		}
		
		// Copy history in backup file
//		boolean renamed = file.renameTo(bakFile);
//		if (!renamed) {
//		    System.out.println("MessageQueue(); failed to rename "
//				       + file.getAbsoluteFile() + " to "
//				       + bakFile.getAbsoluteFile() + ".bak");
//		    return;
//		}
		
//		fout_ =
//		    new PrintStream(new FileOutputStream(file));
//	    
		BufferedReader fin  =
		    new BufferedReader(new FileReader(file));
		
		// Parse backup history and copy message ids in new history
		// file, but only for message that still exist in the directory
		String key = fin.readLine();
		while (key != null) {
		    
//		    if (descriptors_.contains(key)) {
			keys_.add(key);
//			fout_.println(key);
//		    }
		    key = fin.readLine();
		}
		fin.close();
	    }
	    catch(Exception e) {
		e.printStackTrace();
	    }
	}
    }

    // ------------------------------------------------------------
    public void put(Descriptor desc) {
	
	synchronized(lock_) {
	    casa.util.Logger.log("queue", "put(): entering locked section");

	    String key = desc.getDocumentId();
	    
	    if (keys_.contains(key)) {
		System.out.println("MessageQueue.put("
				   + key + ") failed: this message is already in the queue");
		casa.util.Logger.log("queue", "put() failed: duplicate message");
		return;
	    }
	    
	    try {
		byte[] payload = pubSubService_.getAsBytes(key);
		descriptors_.put(key, desc);
		payloads_.put(key, new Payload(payload));
		keys_.add(key);
		fout_.println(key);
		casa.util.Logger.log("queue", "put("
					    + key
					    + ")");
	    } 
	    catch (Exception e) {
		e.printStackTrace();
		return;
	    }

	    casa.util.Logger.log("queue", "put: lock.notify()");
	    lock_.notify();
	}
    }

    // ------------------------------------------------------------
    public void process(Descriptor desc) {
	put(desc);
    }
    
    public File getPath(){
    	return this.path_;
    }
    
    public Repository<String, Descriptor> getDescriptor(){
    	return this.descriptors_;
    }
    public Repository<String, Payload> getPayload(){
    	return this.payloads_;
    }

    // ------------------------------------------------------------
    public Message get(long timeout) {
	
	try {
	    casa.util.Logger.log("queue", "get("
					+ timeout
					+ "): entering locked section");
	    
	    synchronized(lock_) {
		
		if (keys_.isEmpty()) {
		    if (timeout == 0) {
			// We do not want to wait
			casa.util.Logger.log("queue", "get(): returns null");
			return null;
		    }

		    // We are willing to wait...
		    casa.util.Logger.log("queue", "get: lock.wait()");
		    lock_.wait(timeout==-1?0:timeout);
		    casa.util.Logger.log("queue", "get: notified or timeout");
		}
		
		while (! keys_.isEmpty()) {
		    String key = keys_.remove(0);
		    Descriptor desc = descriptors_.get(key);
		    descriptors_.delete(key);
		    Payload payload = payloads_.get(key);
		    payloads_.delete(key);
		    
		    if ((desc != null) && (payload != null)) {
			
			casa.util.Logger.log("queue", "get(): returns " + key);
			return new Message(desc, payload.getBytes());
		    }
		}
		
		casa.util.Logger.log("queue", "get(): returns null");
		return null;
	    }
	}
	catch(InterruptedException e) {
	    System.out.println("Message.get() interrupted");
	    casa.util.Logger.log("queue", "get(): interrupted. Returns null");
	    return null;
	}
    }
    
    //_______ KADER___________________
    public Message get(String key) throws Exception {
    	keys_.remove(key);
	    Descriptor desc = descriptors_.get(key);
	    Payload payload = payloads_.get(key);
	    if ((desc != null) && (payload != null))
	    {
		casa.util.Logger.log("queue", "get(): returns " + key);
		return new Message(desc, payload.getBytes());
		}
	    else
	    	throw new Exception("Message does not exist");
    }
    
    /**
     * delete only this message
     * @param key name of the message
     */
    public void delete(String key) {
    	descriptors_.delete(key);
    	payloads_.delete(key);
    }
    // ------------------------------------------------------------
    public Message get() {
	
	return get(-1);
    }

    // ------------------------------------------------------------
    public int nbMessages() {

	synchronized(lock_) {
	    return descriptors_.size();
	}
    }

    // ------------------------------------------------------------
    public static void main(String[] args)
	throws Exception {

	// Should not work with arg "null" below
	MessageQueue mq = new MessageQueue(new File("queue1"), null);
	
	if (args.length == 0) {
	    Message msg = mq.get();
	    if (msg != null)
		System.out.println(msg.getDescriptor());
	    System.exit(0);
	}

	for (String v: args) {

	    Descriptor desc = new Descriptor();
	    desc.setDocumentId(v);
	    String str = new String("Text is " + v);

	    Message msg = new Message(desc, str.getBytes());
	    System.out.println(msg);
	    
	    mq.put(desc);
	}
    }
}
