/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.util.HashSet;
import java.util.Set;

/**
 * A <code>Dispatcher&lt;E&gt;</code> is a <code>Sink&lt;E&gt;</code> that can
 * feed several <code>Sinks&lt;E&gt;</code> with any object it is written in.
 * <p>
 * Whenever an object is written to a <code>Dispatcher</code>, this object is
 * written in every other sink the dispatcher is associated with.
 * <p>
 * One or several <code>Sinks</code> can be associated with a
 * <code>Dispatcher</code> using method <code>add(Sink)</code>. Sinks can
 * similarly be disassociated from the dispatcher using method
 * <code>remove(Sink)</code>.
 * @author $Author: F. Guidec$
 * @version $Revision: 1.0$
 */

public class Dispatcher<E> implements Sink<E> {
	protected Set<Sink<E>> sinks_ = new HashSet<Sink<E>>();

	public Verbosity verbosity = new Verbosity();

	/**
	 * Creates a new <code>Dispatcher</code>, that is originally associated with
	 * no other <code>Sink</code>
	 */
	public Dispatcher() {
	}

	/*
	 * Associates a new <code>Sink</code> with this <code>Dispatcher</code>
	 * @param sink the <code>Sink</code> that is to be associated with this
	 * <code>Dispatcher</code> from now on
	 */
	public synchronized void addSink(Sink<E> sink) {
		if (verbosity.isEnabled()) {
			System.out.println(this + "Dispatcher.addSink(" + sink + ")");
		}
		sinks_.add(sink);
	}

	/*
	 * Returns the set of all currently associated <code>Sinks</code>
	 * @return the set of all currently associated <code>Sinks</code>.
	 */
	public synchronized Set<Sink<E>> getSinks() {

		if (verbosity.isEnabled()) {
			System.out.println(this + "Dispatcher.getSinks()");
		}

		return sinks_;
	}

	/*
	 * Disassociates a <code>Sink</code> from this <code>Dispatcher</code>
	 * @param sink the <code>Sink</code> that is being disassociated from this
	 * <code>Dispatcher</code>
	 */
	public synchronized void removeSink(Sink<E> sink) {
		if (verbosity.isEnabled()) {
			System.out.println(this + "Dispatcher.removeSink(" + sink + ")");
		}
		sinks_.remove(sink);
	}

	/**
	 * Writes an object in this <code>Dispatcher</code>. This object is then
	 * written in every <code>Sink</code> this dispatcher may be associated with
	 * @param value the object written in this <code>Dispatcher</code>
	 */
	public synchronized void write(E value) {

		if (verbosity.isEnabled()) {
			System.out.println(this + "Dispatcher.write(" + value + ")");
			System.out.println(this + "  Dispatcher has " + sinks_.size()
					+ " client sinks");
		}

		for (Sink<E> sink : sinks_) {
			try {
				if (verbosity.isEnabled())
					System.out.println(this + "  Dispatcher writing to sink "
							+ sink);
				sink.write(value);
			}
			catch (Exception e) {
				if (verbosity.isEnabled()) {
					System.out.println(this
							+ "  Dispatcher failed to write to sink " + sink
							+ ", exception: " + e);
				}
			}
		}
	}

}
