/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.io.File;
import java.io.BufferedReader;
import java.io.PrintStream;
import java.util.Collection;
import java.util.Set;
import java.util.Map;
import java.util.HashMap;
import java.net.InetAddress;

import casa.dodwan.config.DodwanEnvironment;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.message.Message;
import casa.dodwan.message.MessageReceiver;
import casa.dodwan.util.ArgumentNotFoundException;
import casa.dodwan.util.ArgumentParsing;
import casa.dodwan.util.ArgumentParsingException;
import casa.dodwan.util.Command;
import casa.dodwan.util.Console;
import casa.dodwan.util.DataFormatting;
import casa.util.Processor;


public class DomainConsole extends Console {
    
    public Domain domain_;
    
    private Map<String, Descriptor> patterns_ = 
	new HashMap<String, Descriptor>();

    // ------------------------------------------------------------
    public DomainConsole(Domain domain) {

	this("domain%", domain);
}
    
    // ============================================================
    // MAIN COMMANDS
    // ============================================================
    
    // ------------------------------------------------------------
    public DomainConsole(String prompt, Domain domain) {
	
	domain_ = domain;
	prompt_ = prompt;

    	addSubConsole("t", "[t]ransmission",
		      domain_.transmissionService.console());
	
	addSubConsole("g", "[g]ossiping",
		      domain_.gossipingService.console());
    }
    
    // ------------------------------------------------------------
    public void addPattern(String[] arguments, PrintStream out) {

	String key = null;
	String descString = null;
	
	try {
	    key        = ArgumentParsing.searchString("-k", arguments);
	    descString = ArgumentParsing.searchString("-desc", arguments);
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}
	
	Descriptor pattern = Descriptor.fromAttributes(descString);

	patterns_.put(key, pattern);
	domain_.gossipingService.addToLocalProfile(pattern);
    }
    
    // ------------------------------------------------------------
    public void removePattern(String[] arguments, PrintStream out) {
	
	String key = null;
	
	try {
	    key = ArgumentParsing.searchString("-k", arguments);
	} catch (ArgumentNotFoundException e) {
	    Command.missingArgument(e);
	    return;
	} catch (ArgumentParsingException e) {
	    Command.parsingException(e);
	    return;
	} catch (Exception e) {
	    out.println(e);
	}

	Descriptor pattern = patterns_.get(key);

	if (pattern == null) {
	    out.println("Unknown pattern");
	    return;
	}

	patterns_.remove(key);
	domain_.gossipingService.removeFromLocalProfile(pattern);
    }
    
    // ------------------------------------------------------------
    public void listPatterns(String[] arguments, PrintStream out) {

	if (arguments.length != 1) {
	    help("  ", out);
	    return;
	}
	
	out.println(DataFormatting.toString("Id", 10)
		    + "\tPattern");

	for (String id: patterns_.keySet())
	    out.println(DataFormatting.toString(id, 10)
			+ "\t" + patterns_.get(id));
	out.println();
    }

    // ------------------------------------------------------------
    @Override
	public void help(String offset, PrintStream out) {
	out.println(offset + "[a]dd [-k id] [-desc desc_string]\n"
		    + offset + "[l]ist\n"
		    + offset + "[r]emove [-k id]"); 

	subHelp(offset, out);  
    }
    
    // -------------------------------------------------------------
    public void processCommand(String[] arguments, BufferedReader in,
			       PrintStream out) {
	
	if (arguments.length < 1) {
	    help("  ", out);
	    return;
	}
	
	if (arguments[0].startsWith("a")) {
	    addPattern(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("l")) {
	    listPatterns(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("r")) {
	    removePattern(arguments, out);
	    return;
	}
	
	if (arguments[0].startsWith("h")) {
	    out.println("Pub/Sub Service: Command Set\n");
	    help("  ", out);
	    return;
	}
	
  	super.processCommand(arguments, in, out);
    }

}
