/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.util.Set;
import java.util.HashSet;
import java.util.Map;
import java.util.HashMap;
import java.util.Collection;
import java.net.InetSocketAddress;

import casa.util.Logger;
import casa.dodwan.docware.Descriptor;
import casa.dodwan.util.SystemEnvironment;

public class Environment extends Listenable<PresenceListener>
    implements PresenceListener {
    
    private Map<String,Host> peers_ =
	new HashMap<String,Host>();

    private Map<String,Host> oldPeers_ =
	new HashMap<String,Host>();

    private Host localHost_ = new Host(SystemEnvironment.host);
    
    // ============================================================
    // Peer management
    // ============================================================
    
    public void addPeer(String id) {

	synchronized(this) {
	    
	    Host peer = peers_.get(id);
	    if (peer == null) {
		peer = oldPeers_.get(id);
		if (peer == null) 
		    peer = new Host(id);
		else 
		    oldPeers_.remove(id);
		peer.busy.set(false);
		peers_.put(id, peer);
	    }
	    peer.lastAnnounce = System.currentTimeMillis();
	    
	    // Logger.log("peer", "add " + id);
	}
	
	for (PresenceListener listener: getListeners())
	    listener.addPeer(id);
    }
    
    // ------------------------------------------------------------
    public void removePeer(String id) {

	synchronized(this) {
	    Host peer = peers_.get(id);
	    if (peer == null)
		return;
	    
	    oldPeers_.put(id, peer);
	    peers_.remove(id);

	    // long elapsed = System.currentTimeMillis() - peer.lastAnnounce;
	    // Logger.log("peer", "remove " + id
	    // 	       + " (elapsed=" + elapsed + " ms)");
	}
	
	for (PresenceListener listener: getListeners())
	    listener.removePeer(id);
    }
    
    // ------------------------------------------------------------
    public void clearPeers() {

	synchronized(this) {
	    // Logger.log("peer", "clear");
	    
	    oldPeers_.putAll(peers_);
	    peers_.clear();

	}
	for (PresenceListener listener: getListeners())
	    listener.clearPeers();
    }
    
    // ------------------------------------------------------------
    public void updatePeer(String id) {

	synchronized(this) {
	    
	    Host peer = peers_.get(id);
	    if (peer == null)
		return;
	    long now = System.currentTimeMillis();
	    // long elapsed = now - peer.lastAnnounce;
	    peer.lastAnnounce = now;
	    
	    // Logger.log("peer", "update " + id
	    // 	       + " (elapsed=" + elapsed + " ms)");
	}
	
	for (PresenceListener listener: getListeners())
	    listener.updatePeer(id);
    }
    
    // ------------------------------------------------------------
    public synchronized Set<String> getIds() {

	return new HashSet<String>(peers_.keySet());
    }
    
    // ------------------------------------------------------------
    // public synchronized Set<String> getPresenceIds() {
    // 	    Set<String> result = new HashSet<String>();
    // 	    for (Host peer: peers_.values())
    // 		result.add(peer.presenceId);
    // 	    return result;
    // 	}
    
    // ------------------------------------------------------------
    public synchronized Collection<Host> getPeers() {
	return peers_.values();
    }
    
    // ------------------------------------------------------------
    public synchronized boolean isPeer(String id) {
	    return peers_.containsKey(id);
	}
    
    // ------------------------------------------------------------
    public synchronized Host getPeer(String id) {
	    return peers_.get(id);
	}
    
    // ------------------------------------------------------------
    public synchronized void setPeerProfile(String id, 
					    Set<Descriptor> profile) {
	    Host peer = peers_.get(id);
	    if (peer == null) {
		peer = new Host(id);
		peers_.put(id, peer);
	    }
	    peer.profile = profile;
	    
	    // peerModified = true;
	}
    
    // ============================================================
    // Local profile management
    // ============================================================
    
    // ------------------------------------------------------------
    public synchronized Host getLocalHost() {
	    return localHost_;
	}
    
    // ------------------------------------------------------------
    public synchronized Set<Descriptor> getLocalProfile() {
	    return localHost_.profile;
	}
    
}
