/* *****************************************************************************
 * Copyright (c) IRISA Laboratory, 
 * Universite Bretagne Sud, France
 * <http://www-casa.irisa.fr/dodwan>
 *
 * This file is part of DoDWAN.
 * 
 * DoDWAN is free software: you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software 
 * Foundation, either version 3 of the License, or any later
 * version.
 * 
 * DoDWAN is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * DoDWAN.  If not, see <http://www.gnu.org/licenses/>.
 * ****************************************************************************/



package casa.dodwan.util;

import java.util.Date;
import java.util.HashMap;
import java.util.Set;

public class TaskManager<E extends TimerTask> {

	private HashMap<Object, E> tasks_ = new HashMap<Object, E>();

	private Timer timer_;

	public Verbosity verbosity = new Verbosity();

	public TaskManager(String name) {

		timer_ = Timer.getInstance();
	}

	// ------------------------------------------------------------
	public E cancel(Object key) {

		if (verbosity.isEnabled())
			System.out.println("TaskManager.cancel(" + key + ")");

		E task = tasks_.get(key);
		if (task != null) {
			task.cancel();
			tasks_.remove(key);
		}
		return task;
	}

	// ------------------------------------------------------------
	public void clear() {

		if (verbosity.isEnabled())
			System.out.println("TaskManager.clear()");

		for (E task : tasks_.values())
			task.cancel();

		tasks_.clear();
		timer_.purge();
	}

	// ------------------------------------------------------------
	public boolean containsKey(Object key) {

		boolean result = tasks_.containsKey(key);

		if (verbosity.isEnabled())
			System.out.println("TaskManager.containsKey(" + key + "): "
					+ result);

		return result;
	}

	// ------------------------------------------------------------
	public Set<Object> getKeys() {

		return tasks_.keySet();
	}

	// ------------------------------------------------------------
	public E getTask(Object key) {

		E result = tasks_.get(key);

		if (verbosity.isEnabled())
			System.out.println("TaskManager.getTask(" + key + "): " + result);

		return result;
	}

	// ------------------------------------------------------------
	public void schedule(Object key, E task, Date date) throws Exception {

		if (verbosity.isEnabled())
			System.out.println("TaskManager.schedule(" + key + ", " + task
					+ ", " + date + ")");

		if (tasks_.containsKey(key))
			throw new Exception(
					"TaskManager.schedule() failed: This task has already been scheduled");

		tasks_.put(key, task);
		timer_.schedule(task, date);
	}

	// ------------------------------------------------------------
	public void schedule(Object key, E task, long delay) throws Exception {

		if (verbosity.isEnabled())
			System.out.println("TaskManager.schedule(" + key + ", " + task
					+ ", " + delay + ")");

		if (delay <= 0)
			throw new Exception("TaskManager.schedule() failed: Invalid delay");

		if (tasks_.containsKey(key))
			throw new Exception(
					"TaskManager.schedule() failed: This task has already been scheduled");

		tasks_.put(key, task);
		timer_.schedule(task, delay);
	}

	// ------------------------------------------------------------
	public void schedule(Object key, E task, long delay, long period)
			throws Exception {

		if (verbosity.isEnabled())
			System.out.println("TaskManager.schedule(" + key + ", " + task
					+ ", " + delay + ", " + period + ")");

		if (period <= 0)
			throw new Exception("TaskManager.schedule() failed: Invalid period");

		if (tasks_.containsKey(key))
			throw new Exception(
					"TaskManager.schedule() failed: This task has already been scheduled");

		tasks_.put(key, task);
		timer_.schedule(task, delay, period);
	}
}
